#!/usr/bin/env python3

"""Expansion Example

Demonstrates how to expand expressions.
"""

import sympy
from sympy import pprint

def main():
    a = sympy.Symbol('a')
    b = sympy.Symbol('b')
    e = (a + b)**5

    print("\nExpression:")
    pprint(e)
    print('\nExpansion of the above expression:')
    pprint(e.expand())
    print()

if __name__ == "__main__":
    main()
